/*******************************************************************************
 * Copyright (c) 2000, 2008 IBM Corporation and others.
 * Portion Copyright (c) 2009-2010 compeople AG (http://www.compeople.de).
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *     Compeople AG	- QtJambi/Qt based implementation for Windows/Mac OS X/Linux
 *******************************************************************************/
package org.eclipse.swt.widgets;

import com.trolltech.qt.core.QDir;
import com.trolltech.qt.gui.QFileDialog;

import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTException;

/**
 * Instances of this class allow the user to navigate the file system and select
 * a directory.
 * <dl>
 * <dt><b>Styles:</b></dt>
 * <dd>(none)</dd>
 * <dt><b>Events:</b></dt>
 * <dd>(none)</dd>
 * </dl>
 * <p>
 * IMPORTANT: This class is intended to be subclassed <em>only</em> within the
 * SWT implementation.
 * </p>
 * 
 * @see <a
 *      href="http://www.eclipse.org/swt/snippets/#directorydialog">DirectoryDialog
 *      snippets</a>
 * @see <a href="http://www.eclipse.org/swt/examples.php">SWT Example:
 *      ControlExample, Dialog tab</a>
 * @see <a href="http://www.eclipse.org/swt/">Sample code and further
 *      information</a>
 * @noextend This class is not intended to be subclassed by clients.
 */

public class DirectoryDialog extends Dialog {
	String message = "", filterPath = ""; //$NON-NLS-1$//$NON-NLS-2$
	String directoryPath = ""; //$NON-NLS-1$
	String[] dirNames = new String[0];
	QFileDialog fd;

	/**
	 * Constructs a new instance of this class given only its parent.
	 * 
	 * @param parent
	 *            a shell which will be the parent of the new instance
	 * 
	 * @exception IllegalArgumentException
	 *                <ul>
	 *                <li>ERROR_NULL_ARGUMENT - if the parent is null</li>
	 *                </ul>
	 * @exception SWTException
	 *                <ul>
	 *                <li>ERROR_THREAD_INVALID_ACCESS - if not called from the
	 *                thread that created the parent</li>
	 *                <li>ERROR_INVALID_SUBCLASS - if this class is not an
	 *                allowed subclass</li>
	 *                </ul>
	 */
	public DirectoryDialog(Shell parent) {
		this(parent, SWT.APPLICATION_MODAL);
	}

	/**
	 * Constructs a new instance of this class given its parent and a style
	 * value describing its behavior and appearance.
	 * <p>
	 * The style value is either one of the style constants defined in class
	 * <code>SWT</code> which is applicable to instances of this class, or must
	 * be built by <em>bitwise OR</em>'ing together (that is, using the
	 * <code>int</code> "|" operator) two or more of those <code>SWT</code>
	 * style constants. The class description lists the style constants that are
	 * applicable to the class. Style bits are also inherited from superclasses.
	 * </p>
	 * 
	 * @param parent
	 *            a shell which will be the parent of the new instance
	 * @param style
	 *            the style of dialog to construct
	 * 
	 * @exception IllegalArgumentException
	 *                <ul>
	 *                <li>ERROR_NULL_ARGUMENT - if the parent is null</li>
	 *                </ul>
	 * @exception SWTException
	 *                <ul>
	 *                <li>ERROR_THREAD_INVALID_ACCESS - if not called from the
	 *                thread that created the parent</li>
	 *                <li>ERROR_INVALID_SUBCLASS - if this class is not an
	 *                allowed subclass</li>
	 *                </ul>
	 */
	public DirectoryDialog(Shell parent, int style) {
		super(parent, checkStyle(parent, style));
		checkSubclass();
		fd = new QFileDialog();
		fd.setFileMode(QFileDialog.FileMode.DirectoryOnly);
	}

	/**
	 * Returns the path which the dialog will use to filter the directories it
	 * shows.
	 * 
	 * @return the filter path
	 * 
	 * @see #setFilterPath
	 */
	public String getFilterPath() {
		QDir dir = fd.directory();
		return dir.absolutePath();
	}

	/**
	 * Returns the dialog's message, which is a description of the purpose for
	 * which it was opened. This message will be visible on the dialog while it
	 * is open.
	 * 
	 * @return the message
	 */
	public String getMessage() {
		return message;
	}

	/**
	 * Makes the dialog visible and brings it to the front of the display.
	 * 
	 * @return a string describing the absolute path of the selected directory,
	 *         or null if the dialog was cancelled or an error occurred
	 * 
	 * @exception SWTException
	 *                <ul>
	 *                <li>ERROR_WIDGET_DISPOSED - if the dialog has been
	 *                disposed</li>
	 *                <li>ERROR_THREAD_INVALID_ACCESS - if not called from the
	 *                thread that created the dialog</li>
	 *                </ul>
	 */
	public String open() {

		java.util.List<String> dirs = null;
		if (fd.exec() != 0) {
			dirs = fd.selectedFiles();
		}
		directoryPath = dirs.get(0); // 
		fd.disposeLater();

		return directoryPath;
	}

	/**
	 * Sets the path that the dialog will use to filter the directories it shows
	 * to the argument, which may be null. If the string is null, then the
	 * operating system's default filter path will be used.
	 * <p>
	 * Note that the path string is platform dependent. For convenience, either
	 * '/' or '\' can be used as a path separator.
	 * </p>
	 * 
	 * @param string
	 *            the filter path
	 */
	public void setFilterPath(String string) {
		fd.setDirectory(string);
	}

	/**
	 * Sets the dialog's message, which is a description of the purpose for
	 * which it was opened. This message will be visible on the dialog while it
	 * is open.
	 * 
	 * @param string
	 *            the message
	 * 
	 * @exception IllegalArgumentException
	 *                <ul>
	 *                <li>ERROR_NULL_ARGUMENT - if the string is null</li>
	 *                </ul>
	 */
	public void setMessage(String string) {
		/*
		 * QFileDialog has nothing like this message. There's a caption which
		 * will be shown in the title of the dialog window. But this has to be
		 * set when the dialog is initialized. There are also some labelTexts
		 * which can be set like setLabelText(DialogLabel.FileName, "Choose:")
		 * but these are only for describing buttons and textfields
		 */
		if (string == null) {
			error(SWT.ERROR_NULL_ARGUMENT);
		}

		message = string;
	}
}
